-module(sat).
-compile(export_all).

benchmark(F) -> 
    {ok, Clauses} = file:consult("formula.txt"),
    timer:tc(?MODULE, F, Clauses).

solve([]) ->
    [];
solve([[]|_]) ->
    false;
solve([[L|Lits]|Clauses]) ->
    case solve(unit_propagation(L,Clauses)) of
	false ->
	    case solve(unit_propagation(-L,[Lits|Clauses])) of
		false ->
		    false;
		Solution ->
		    [-L|Solution]
	    end;
	Solution ->
	    [L|Solution]
    end.

unit_propagation(L,Clauses) ->
    NewClauses =
	[lists:delete(-L,C) || C <- Clauses,
			       not lists:member(L,C)],
    sort_by_length(NewClauses).

sort_by_length(Clauses) ->
    [C || {_,C} <- lists:usort([{length(C),C} || C <- Clauses])].

par_solve([]) ->
    [];
par_solve([[]|_]) ->
    false;
par_solve([[L|Lits]|Clauses]) ->
    Rest = speculate(fun()-> par_solve(unit_propagation(-L,[Lits|Clauses])) end),
    case par_solve(unit_propagation(L,Clauses)) of
	false ->
	    case value_of(Rest) of
		false ->
		    false;
		Solution ->
		    [-L|Solution]
	    end;
	Solution ->
	    [L|Solution]
    end.

speculate(F) ->
    Parent = self(),
    Pid = spawn_link(fun() -> Parent ! {self(),F()} end),
    {speculating,Pid}.

value_of({speculating,Pid}) ->
    receive {Pid,X} ->
	    X
    end.


%% Worker pools
start_pool(N) ->
    true = register(pool,spawn_link(fun()->pool([worker() || _ <- lists:seq(1,N)]) end)).

pool(Workers) ->
    pool(Workers,Workers).

pool(Workers,All) ->
    receive
	{get_worker,Pid} ->
	    case Workers of
		[] ->
		    Pid ! {pool,no_worker},
		    pool(Workers,All);
		[W|Ws] ->
		    Pid ! {pool,W},
		    pool(Ws,All)
	    end;
	{return_worker,W} ->
	    pool([W|Workers],All);
	{stop,Pid} ->
	    [unlink(W) || W <- All],
	    [exit(W,kill) || W <- All],
	    unregister(pool),
	    Pid ! {pool,stopped}
    end.

worker() ->
    spawn_link(fun work/0).

work() ->
    receive
	{task,Pid,R,F} ->
	    Pid ! {R,F()},
	    pool ! {return_worker,self()},
	    work()
    end.

speculate_on_worker(F) ->
    case whereis(pool) of
	undefined ->
	    ok; %% we're stopping
	Pool -> Pool ! {get_worker,self()}
    end,
    receive
	{pool,no_worker} ->
	    {not_speculating,F};
	{pool,W} ->
	    R = make_ref(),
	    W ! {task,self(),R,F},
	    {speculating,R}
    end.

worker_value_of({not_speculating,F}) ->
    F();
worker_value_of({speculating,R}) ->
    receive
	{R,X} ->
	    X
    end.


pool_solve(P) ->
    start_pool(erlang:system_info(schedulers)-1),
    S = pool_solve1(P),
    pool ! {stop,self()},
    receive {pool,stopped} -> S end.

pool_solve1([]) ->
    [];
pool_solve1([[]|_]) ->
    false;
pool_solve1([[L|Lits]|Clauses]) ->
    Rest = speculate_on_worker(
	     fun()-> pool_solve1(unit_propagation(-L,[Lits|Clauses])) end),
    case pool_solve1(unit_propagation(L,Clauses)) of
	false ->
	    case worker_value_of(Rest) of
		false ->
		    false;
		Solution ->
		    [-L|Solution]
	    end;
	Solution ->
	    [L|Solution]
    end.

limit_solve(P) ->
    start_pool(erlang:system_info(schedulers)-1),
    S = limit_solve1(P),
    pool ! {stop,self()},
    receive {pool,stopped} -> S end.

limit_solve1([]) ->
    [];
limit_solve1([[]|_]) ->
    false;
limit_solve1([[L|Lits]|Clauses]) when length(Clauses) < 250 ->
    solve([[L|Lits]|Clauses]);
limit_solve1([[L|Lits]|Clauses]) ->
    Rest = speculate_on_worker(
	     fun()-> limit_solve1(unit_propagation(-L,[Lits|Clauses])) end),
    case limit_solve1(unit_propagation(L,Clauses)) of
	false ->
	    case worker_value_of(Rest) of
		false ->
		    false;
		Solution ->
		    [-L|Solution]
	    end;
	Solution ->
	    [L|Solution]
    end.
