import Data.List
import Data.Function
import Control.Parallel

-- | A variable in a SAT problem is a strictly positive integer; @1@ denotes
--   variable @x1@, 42 denotes @x42@, etc.
--   A literal is a variable which may or may not be negated.
--   Variables are negated by taking their negation.
type Lit = Int

-- | A clause is a disjunction of literals. An empty clause is unsatisfiable.
type Clause = [Lit]

-- | A problem is a conjunction of literals. An empty problem is valid.
--   Invariant: clauses are kept sorted by length.
type Problem = [Clause]

-- | A solution is a list of literals.
type Solution = [Lit]

-- | Force the given literal. All clauses where the literal appears are removed
--   as they are by definition satisfied under this assignment.
--   The negation of the literal is removed from all remaining clauses, since
--   it is not by definition unsatisfiable.
--
--   Maintains the invariant that clauses are sorted by length.
force :: Lit -> Problem -> Problem
force l clauses =
  sortBy (compare `on` length) [delete (-l) c | c <- clauses, not $ l `elem` c]

-- | Solve a SAT problem.
--
--   If the problem contains no clauses, it is valid and an empty solution
--   is returned to signify this fact.
--
--   If any clause is empty it is unsatisfiable, which means the entire problem
--   is unsatisfiable; @Nothing@ is returned.
--
--   Otherwise, we pick a literal from the shortest clause, propagate it
--   throughout the problem, and try to solve the problem recursively.
--   In this case, the clause the literal was picked from is not included when
--   solving recursively, as it is already satisfied by virtue of containing
--   the literal in the first place.
--
--   If no solution is found this way, we instead force the negation of the
--   chosen literal and solve recursively. Note that in this case, the clause
--   from which the literal was chosen IS included in the recursive solution,
--   as it is not yet satisfied, but may still be satisfiable.
solve :: Problem -> Maybe Solution
solve ((l:ls):cs)
  | Just sol <- solve (force l cs)           = Just $ l : sol
  | Just sol <- solve (force (-l) (ls : cs)) = Just $ (-l) : sol
  | otherwise                                = Nothing
solve []                                     = Just []
solve ([] : _)                               = Nothing

-- | Overly naïve parallel version of 'solve'.
--   Still significantly faster though.
solve' :: Problem -> Maybe Solution
solve' ((l:ls):cs) =
    case pos `par` neg `pseq` (pos, neg) of
      (Just sol, _) -> Just sol
      (_, Just sol) -> Just sol
      _             -> Nothing
  where
    pos = solve' (force l cs)
    neg = solve' (force (-l) (ls : cs))
solve' []                                     = Just []
solve' ([] : _)                               = Nothing

-- | The problem from the exercise session.
problem :: Problem
problem =
  [ [10,-3,68]
  , [-23,-62,42]
  , [-15,-1,-18]
  , [-19,51,60]
  , [38,15,-56]
  , [-49,-48,55]
  , [28,16,61]
  , [35,66,-40]
  , [25,30,-55]
  , [79,-73,50]
  , [69, 60, 71]
  , [-9,21,66]
  , [27,-14,-5]
  , [-39,61,26]
  , [35,-44,15]
  , [53,-41,68]
  , [-9,67,78]
  , [-66,-6,8]
  , [-40,-30,-52]
  , [-24,-8,29]
  , [-23,80,42]
  , [75,-66,-18]
  , [-35,-55,60]
  , [39,62,-13]
  , [-70,37,61]
  , [-21,-20,12]
  , [-79,24,-51]
  , [9,-42,74]
  , [57,45,-32]
  , [67,-22,21]
  , [70,-14,-3]
  , [60,-58,34]
  , [74,-24,-41]
  , [8,-31,-21]
  , [18,-17,22]
  , [-20,61,37]
  , [-76,-53,-55]
  , [-19,2,35]
  , [14,27,72]
  , [54,63,-62]
  , [6,52,-61]
  , [-62,-50,-13]
  , [-12,-69,-9]
  , [-40,70,-14]
  , [-76,-10,28]
  , [-19,29,-72]
  , [-80,-35,1]
  , [-38,-80,-46]
  , [75,-51,-19]
  , [-72,-67,-26]
  , [-24,-8,-36]
  , [-16,47,-33]
  , [37,59,-62]
  , [76,53,-45]
  , [-17,80,56]
  , [-8,-48,-40]
  , [-60,22,-58]
  , [-32,30,66]
  , [8,-55,21]
  , [42,45,26]
  , [11,53,-1]
  , [-3,14,-16]
  , [57,-38,-75]
  , [43, 66, 56]
  , [78,22,-37]
  , [-39,5,67]
  , [-75,-53,63]
  , [52,5,29]
  , [34,-31,-63]
  , [53,-33,36]
  , [-39,70,69]
  , [-34,-59,-33]
  , [37,-5,49]
  , [20,-52,-17]
  , [-66,46,-14]
  , [-74,-72,56]
  , [76,-4,-46]
  , [64,-47,13]
  , [76,56,22]
  , [-40,-16,-12]
  , [-77,-65,-53]
  , [30,70,8]
  , [72,26,-34]
  , [63,1,24]
  , [10,-38,-25]
  , [54,-6,-10]
  , [63,11,-32]
  , [55,-71,-25]
  , [78,-15,21]
  , [-62,60,-59]
  , [20,51,-25]
  , [-24,-58,62]
  , [31,77,47]
  , [71,74,26]
  , [-70,-78,8]
  , [39,-26,77]
  , [78,-53,68]
  , [-48,63,-76]
  , [24,77,-71]
  , [1,72,23]
  , [-55,-40,70]
  , [-55,-79,-15]
  , [-58,76,80]
  , [68,17,22]
  , [-19,-52,27]
  , [-66,-12,40]
  , [-28,23,-32]
  , [-74,-16,3]
  , [-22,49,-70]
  , [-2,62,26]
  , [-38,-67,-30]
  , [-45,79,-36]
  , [-29,-47,-17]
  , [51,10,-79]
  , [-67,15,-79]
  , [-26,-7,32]
  , [51,2,-59]
  , [5,-64,42]
  , [-65,-49,-35]
  , [-1,-42,-79]
  , [22,-1,43]
  , [-74,31,10]
  , [58,73,-30]
  , [-26,22,-45]
  , [-78,25,-38]
  , [-27,-46,9]
  , [-12,49,38]
  , [-19,-25,53]
  , [-34,-5,-75]
  , [54,35,-65]
  , [8,29,14]
  , [-24,69,-4]
  , [-35,21,-38]
  , [27,73,-38]
  , [-20,29,80]
  , [-70,71,-6]
  , [-55,46,-3]
  , [37,-65,-46]
  , [16,56,-70]
  , [24,-43,14]
  , [49,-39,-22]
  , [60,23,-63]
  , [32,27,3]
  , [76,14,20]
  , [43,-55,-11]
  , [-47,5,-16]
  , [-77,-24,-58]
  , [34,-16,-77]
  , [-56,-73,-18]
  , [-74,-40,-78]
  , [-40,-61,26]
  , [-46,39,-29]
  , [72,74,-4]
  , [32,-14,36]
  , [50,-61,54]
  , [-27,-66,59]
  , [-41,69,-9]
  , [-1,28,6]
  , [6,-5,-25]
  , [-70,-53,36]
  , [-37,-74,-16]
  , [71, 66, 11]
  , [32,43,-37]
  , [25,72,-22]
  , [-12,74,-61]
  , [59,56,18]
  , [10,36,-21]
  , [-47,33,-46]
  , [-60,13,-42]
  , [-11,18,-23]
  , [69,-43,51]
  , [-80,1,-54]
  , [-54,-35,-55]
  , [-25,10,-30]
  , [-48,-66,3]
  , [-10,26,-14]
  , [-45,54,49]
  , [-49,-57,20]
  , [-80,65,-32]
  , [33,-80,-18]
  , [-44,10,-4]
  , [-56,-13,21]
  , [57,47,-28]
  , [26,-19,-1]
  , [-72,34,28]
  , [32, 71, 61]
  , [-27,12,13]
  , [-20,5,-48]
  , [-51,-79,16]
  , [80,20,56]
  , [19,-60,71]
  , [-78,-43,-33]
  , [-25,-22,-34]
  , [-22,76,55]
  , [-6,-26,63]
  , [-57,43,-22]
  , [-34,-67,-47]
  , [29,43,-40]
  , [68,-10,-43]
  , [38,-60,2]
  , [70, 75, 64]
  , [5,-26,75]
  , [9,-34,-55]
  , [-61,76,-64]
  , [-34,-35,72]
  , [33, 69, 32]
  , [-54,-12,57]
  , [-27,56,68]
  , [-38,-43,-13]
  , [68,-49,9]
  , [-63,-73,70]
  , [61,60,1]
  , [57,46,-24]
  , [-62,-19,37]
  , [3,7,68]
  , [51,-52,22]
  , [-32,-21,-45]
  , [-11,-56,-8]
  , [-41,-12,-15]
  , [-62,24,13]
  , [-26,43,-63]
  , [3,30,-57]
  , [-30,38,-63]
  , [5,-23,-34]
  , [-46,-2,13]
  , [-58,61,35]
  , [16,-56,-41]
  , [20,48,76]
  , [-28,58,14]
  , [-44,-55,52]
  , [79,12,-32]
  , [13,19,73]
  , [57,52,-4]
  , [17,18,-21]
  , [51, 57, 39]
  , [72,13,-29]
  , [53, 63, 71]
  , [-51,52,49]
  , [-48,-9,-40]
  , [41,-62,59]
  , [43,-60,-6]
  , [-58,-48,26]
  , [48,30,-76]
  , [-73,66,6]
  , [-56,-75,18]
  , [-32,27,17]
  , [-12,-23,-68]
  , [-55,50,25]
  , [53,-68,-7]
  , [27,-28,72]
  , [28,80,3]
  , [-58,-57,-50]
  , [-42,-79,51]
  , [-69,-6,71]
  , [-32,-26,16]
  , [-36,-33,9]
  , [75,47,-44]
  , [-55,73,33]
  , [19,35,-60]
  , [19,57,-32]
  , [-7,10,-2]
  , [-38,-49,-3]
  , [-20,-66,-60]
  , [-13,43,79]
  , [-65,69,42]
  , [-33,58,77]
  , [-25,-35,-77]
  , [61,-51,1]
  , [-60,-19,-49]
  , [28,18,-43]
  , [-63,-9,-47]
  , [42,-40,-15]
  , [-36,-2,79]
  , [66,17,24]
  , [75,-1,17]
  , [25,-24,-16]
  , [70,-67,48]
  , [70,-57,78]
  , [11,-47,-78]
  , [-3,63,-66]
  , [45,63,-54]
  , [-15,61,9]
  , [-65,-29,-16]
  , [-75,-7,-80]
  , [75,-50,36]
  , [-29,20,33]
  , [-5,67,-56]
  , [61,44,-27]
  , [-64,-25,-57]
  , [-66,15,65]
  , [-70,-65,51]
  , [-76,61,13]
  , [-25,-54,-28]
  , [39,22,-9]
  , [25,-53,-74]
  , [-26,41,47]
  , [47,41,25]
  , [53,43,-11]
  , [-10,-18,15]
  , [58, 63, 47]
  , [-2,-71,73]
  , [64,-40,24]
  , [-71,-63,-48]
  , [76,-80,44]
  , [4,-62,-51]
  , [66,21,72]
  , [-8,-24,19]
  , [-38,-63,35]
  , [23,-29,-4]
  , [-9,22,48]
  , [5,-32,52]
  , [-47,21,-7]
  , [59, 45, 65]
  , [29,71,57]
  , [27,78,-77]
  , [20,-5,25]
  , [-75,-5,-26]
  , [-49,-42,-56]
  , [-10,-4,-43]
  , [-50,-19,5]
  , [68,36,-31]
  , [-52,47,-9]
  , [51,-52,-19]
  , [34,-58,-27]
  , [-62,72,9]
  , [-46,47,21]
  , [-46,11,-52]
  , [14,10,12]
  , [78,33,-27]
  , [-7,1,5]
  , [11,63,-45]
  , [26,66,56]
  , [-34,-14,-72]
  , [-41,-13,1]
  , [1,20,40]
  , [73,-8,6]
  , [22,-48,33]
  , [35,19,51]
  , [30,-23,42]
  , [-71,-77,56]
  , [62,33,-23]
  , [5,-70,34]
  , [-4,-15,-2]
  , [72,38,-29]
  ]

main = print $ solve' problem
